import bs
import weakref
import os
import thread
import threading
import time
import random
import urllib
import urllib2
import ast
import gc
import types
import copy
import _bs

_gPlayerColors = [(1,0.25,0.2),(0.2,1,0.2),(0.1,0.25,1),(0.2,1,1),
                  (0.65,0.1,1.0),(1,1,0),(1,0.5,0),(1,0.3,0.5),
                  (0.1,0.1,0.5),(0.4,0.2,0.1),(0.1,0.35,0.1),(1,0.8,0.5),
                  (0.4,0.05,0.05),(0.13,0.13,0.13),(0.5,0.5,0.5),(1,1,1)]
gMusic = None

gNodeOwnerWeakRefs = []
gNodeOwnerWeakRefsCleanCounter = 0

def showUserScripts():
    "Open the user scripts directory in a filesystem browser (on applicable systems)"
    ua = bs.getEnvironment()['userAgentString']
    if 'Mac' in ua: os.system("open '"+bs.getEnvironment()['userScriptsDirectory']+"'")
    else: bs.screenMessage(bs.getEnvironment()['userScriptsDirectory'])

def createUserSystemScripts():
    'Sets up a copy of BombSquad system scripts under your user scripts dir that you can edit and experiment with.'
    import shutil
    path = bs.getEnvironment()['userScriptsDirectory']+'/sys/'+bs.getEnvironment()['version']
    if os.path.exists(path): shutil.rmtree(path)
    if os.path.exists(path+"_tmp"): shutil.rmtree(path+"_tmp")

    try:
        os.makedirs(path+'_tmp')
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path+"_tmp"): pass
        else: raise

    # shutil.copytree doesn't seem to work nicely on android, so lets do it manually
    srcDir = bs.getEnvironment()['systemScriptsDirectory']
    dstDir = path+"_tmp"
    files = os.listdir(bs.getEnvironment()['systemScriptsDirectory'])
    for f in files:
        print 'COPYING',srcDir+'/'+f,'->',dstDir
        shutil.copyfile(srcDir+'/'+f,dstDir+'/'+f)

    print 'MOVING',path+"_tmp",path
    shutil.move(path+"_tmp",path)
    #if 'android' in bs.getEnvironment()['userAgentString']: bs.androidRefreshFiles() # get android file transfer to reflect this
    print 'Created system scripts at :\''+path+'\'\nRestart BombSquad to use them. (use bs.quit() to exit the game) '

def deleteUserSystemScripts():
    'Cleans out the scripts created by createUserSystemScripts()'
    import shutil
    path = bs.getEnvironment()['userScriptsDirectory']+'/sys/'+bs.getEnvironment()['version']
    if os.path.exists(path):
        shutil.rmtree(path)
        print 'User system scripts deleted.\nRestart BombSquad to use internal scripts. (use bs.quit() to exit the game)'
    else:
        print 'User system scripts not found.'

    # if the sys path is empty, kill it
    d = bs.getEnvironment()['userScriptsDirectory']+'/sys'
    if os.path.isdir(d) and len(os.listdir(d)) == 0:
        os.rmdir(d)
    if 'android' in bs.getEnvironment()['userAgentString']: bs.androidRefreshFiles() # get android file transfer to reflect this


def getTypeName(t):
    return t.__module__+'.'+t.__name__

def resolveTypeName(typeName):
    moduleName,className = typeName.split('.')
    module = __import__(moduleName)
    return getattr(module,className)
    
def getPlayerColors():
    """ return user-selectable player colors """
    return _gPlayerColors

# call suppressDebugReports() to keep from spamming the debug log server during development
_gSuppressDebugReports = False
def suppressDebugReports():
    bs.screenMessage("Supressing debug reports",color=(1,0,0))
    global _gSuppressDebugReports
    _gSuppressDebugReports = True
_gPrintedLiveObjectWarning = False

def animate(node,attr,keys,loop=False,offset=0,driver='gameTime'):
    """Create an animCurve node with the provided values and time as an input,
    connect it to the provided attribute, and set it to die with the target.
    Key values are provided as time:value dictionary pairs.  Time values are relative
    to the current time. Returns the nodeID of the animCurve"""

    if driver != 'gameTime': raise Exception("fixme; only support game-time currently")
    items = keys.items()
    items.sort()
    curve = bs.newNode("animCurve",owner=node,name='Driving '+str(node)+' \''+attr+'\'')
    bs.time.connectAttr(driver,curve,"in")
    curve.connectAttr("out",node,attr)
    curve.times = [time for time,val in items]
    curve.values = [val for time,val in items]
    curve.loop = loop
    curve.offset = bs.getGameTime()+offset
    # if we're not looping, set a timer to kill this curve after its done its job
    # FIXME - even if we are looping we should have a way to die once we get disconnected
    if loop == False:
        bs.gameTimer(int(items[-1][0])+1000,curve.delete)
    return curve

def animateArray(node,attr,size,keys,loop=False,offset=0,driver='gameTime'):
    combine = bs.newNode('combine',owner=node,attrs={'size':size})
    if driver != 'gameTime': raise Exception("fixme; only support game-time currently")
    items = keys.items()
    items.sort()
    curves = {}
    for i in range(size):
        curve = bs.newNode("animCurve",owner=node,name='Driving '+str(node)+' \''+attr+'\' member '+str(i))
        bs.time.connectAttr(driver,curve,"in")
        curve.connectAttr("out",combine,'input'+str(i))
        curve.times = [time for time,val in items]
        curve.values = [val[i] for time,val in items]
        curve.loop = loop
        curve.offset = bs.getGameTime()+offset
        # if we're not looping, set a timer to kill this curve after its done its job
        if loop == False:
            bs.gameTimer(int(items[-1][0])+1000,curve.delete)
    combine.connectAttr('output',node,attr)
    # if we're not looping, set a timer to kill the combine once the job is done
    # FIXME - even if we are looping we should have a way to die once we get disconnected
    if loop == False:
        bs.gameTimer(int(items[-1][0])+1000,combine.delete)



# called internally..
def _shutdown():
    if _gMusicPlayer is not None:
        _gMusicPlayer.shutdown()
    # if _gITunesThread is not None:
    #     print 'CALLING SHUTDOWN'
    #     _gITunesThread.shutdown()



class MusicPlayer(object):

    def __init__(self):
        self._haveSetInitialVolume = False
        self._entryToPlay = None
        self._volume = 1.0
        self._actuallyPlaying = False

    def selectEntry(self,callback,currentEntry,selectionTargetName):
        return self.onSelectEntry(callback,currentEntry,selectionTargetName)
        
    def setVolume(self,volume):
        oldVolume = self._volume
        self._volume = volume
        self.onSetVolume(volume)
        self._updatePlayState()

    def _updatePlayState(self):
        # if we arent playing, should be, and have positive volume, do so
        if not self._actuallyPlaying:
            if self._entryToPlay is not None and self._volume > 0.0:
                self.onPlay(self._entryToPlay)
                self._actuallyPlaying = True
        else:
            if self._actuallyPlaying and (self._entryToPlay is None or self._volume <= 0.0):
                self.onStop()
                self._actuallyPlaying = False
                
    def play(self,entry):
        if not self._haveSetInitialVolume:
            self._volume = _bs._getSetting('Music Volume')
            self.onSetVolume(self._volume)
            self._haveSetInitialVolume = True
        self._entryToPlay = copy.deepcopy(entry)

        # if we're currently *actually* playing something, switch to the new thing..
        # otherwise update state which will start us playing *only* if proper (volume > 0, etc)
        if self._actuallyPlaying: self.onPlay(self._entryToPlay)
        else: self._updatePlayState()

    def stop(self):
        self._entryToPlay = None
        self._updatePlayState()

    def shutdown(self):
        self.onShutdown()
        
    # subclasses should override the following:
    def onSelectEntry(self,callback,currentEntry,selectionTargetName):
        'should present a GUI to select an entry; callback should be called with an entry or None for default'
        pass
    
    def onSetVolume(self,volume):
        'Called when the volume should be changed'
        pass

    def onPlay(self,entry):
        'Called when a new song/playlist/etc should be played'
        pass

    def onStop(self):
        'Called when the music should stop'
        pass

    def onShutdown(self):
        'Called on final app shutdown'
        pass



_gMusicPlayer = None
_gMusicPlayerType = None
#_gITunesThread = None

# for internal music player
def _getValidMusicFileExtensions():
    return ['mp3','ogg','m4a','wav','flac','mid']

if 'android' in bs.getEnvironment()['userAgentString']:

    # this talks to native stuff via our c layer
    class InternalMusicPlayer(MusicPlayer):
        def __init__(self):
            MusicPlayer.__init__(self)
            self._wantToPlay = False
            self._actuallyPlaying = False
            
        def onSelectEntry(self,callback,currentEntry,selectionTargetName):
            import bsUI
            return bsUI.GetSoundtrackEntryTypeWindow(callback,currentEntry,selectionTargetName)
            
        def onSetVolume(self,volume):
            _bs._musicPlayerSetVolume(volume);

        class _PickFolderSongThread(threading.Thread):
            def __init__(self,path,callback):
                threading.Thread.__init__(self)
                self._callback = callback
                self._path = path

            def run(self):
                try:
                    allFiles = []
                    validExtensions = ['.'+x for x in _getValidMusicFileExtensions()]
                    for root, subFolders, files in os.walk(self._path):
                        for f in files:
                            if any(f.lower().endswith(ext) for ext in validExtensions):
                                allFiles.insert(random.randrange(len(allFiles)+1),root+'/'+f)
                    if len(allFiles) == 0: raise Exception(bs.getResource('internal.noMusicFilesInFolderText'))
                    #bs.runInGameThread(bs.Call(self._callback,fileName=allFiles[random.randrange(len(allFiles))]))
                    bs.runInGameThread(bs.Call(self._callback,result=allFiles))
                except Exception,e:
                    bs.runInGameThread(bs.Call(self._callback,result=self._path,error=str(e)))
            
        def onPlay(self,entry):
            entryType = _getSoundtrackEntryType(entry)
            name = _getSoundtrackEntryName(entry)
            if entryType == 'musicFile':
                self._wantToPlay = self._actuallyPlaying = True
                _bs._musicPlayerPlay(name)
            elif entryType == 'musicFolder':
                # launch a thread to scan this folder and give us a random valid file within
                self._wantToPlay = True
                self._actuallyPlaying = False
                self._PickFolderSongThread(name,self._onPlayFolderCB).start()

        def _onPlayFolderCB(self,result,error=None):
            if error is not None:
                errStr = bs.getResource('internal.errorPlayingMusicText').replace('${MUSIC}',os.path.basename(result))+'; '+str(error)
                #print errStr
                bs.screenMessage(errStr,color=(1,0,0))
                return

            # theres a chance a stop could have been issued before our thread returned..
            # if that's the case, dont' play.
            if not self._wantToPlay:
                print '_onPlayFolderCB called with _wantToPlay False'
            else:
                self._actuallyPlaying = True
                _bs._musicPlayerPlay(result)
        
        def onStop(self):
            self._wantToPlay = False
            self._actuallyPlaying = False
            _bs._musicPlayerStop()

        def onShutdown(self):
            _bs._musicPlayerShutdown()

    _gMusicPlayerType = InternalMusicPlayer

elif 'Mac' in bs.getEnvironment()['userAgentString']:

    class ITunesThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self._commandsAvailable = threading.Event()
            self._commands = []
            self._volume = 1.0
            self._currentPlaylist = None

        def run(self):
            import ScriptingBridge as sb
            import LaunchServices as ls

            # iTunes seems to be locking up sometimes when bombsquad quits.
            # I'm guessing it might be related to the scripting bridge app going down
            # when this thread dies.
            # ..so as a workaround, lets make it a global so it won't get killed
            global gITunesApp
            try: gITunesApp
            except Exception:
                self._app = gITunesApp = sb.SBApplication.applicationWithBundleIdentifier_("com.apple.iTunes")
                # it looks like launching iTunes here on 10.7/10.8 knocks us out of fullscreen.
                # ew... that might be a bug, but for now we can work around it by reactivating
                # ourself after
                def doPrint():
                    bs.realTimer(1000,bs.Call(bs.screenMessage,bs.getResource('usingItunesText'),(0,1,0)))
                    bs.realTimer(3000,bs.Call(bs.screenMessage,bs.getResource('usingItunesTurnRepeatAndShuffleOnText'),(0,1,0)))

                bs.runInGameThread(doPrint)
                # here we grab this to force the actual launch
                # ..currently (on 10.8 at least) this is causing a switch away from our fullscreen window.
                # to work around this we explicitly focus our main window to bring ourself back
                self._app.soundVolume()
                bs.runInGameThread(_bs._focusWindow)

            self._source = self._app.sources().objectWithName_('Library')
            if not self._source.exists():
                self._source = self._app.sources()[0] # as a fallback just grab first
            done = False
            while not done:
                self._commandsAvailable.wait()
                self._commandsAvailable.clear()
                # (we're not protecting this list with a mutex but we're
                # just using it as a simple queue so it should be fine..)
                while len(self._commands) > 0:
                    cmd = self._commands.pop(0)
                    if cmd[0] == 'DIE':
                        # only stop if we've actually played something..
                        # (we don't want to be killing music the user has playing)
                        #if self._havePlayed:
                        if self._currentPlaylist and self._volume > 0:
                            try:
                                self._app.stop()
                                self._app.setSoundVolume_(self._origVolume)
                            except Exception,e: print 'Error stopping iTunes music:',e
                        done = True
                        break
                    elif cmd[0] == 'PLAY':
                        # None implies stop
                        if cmd[1] == None:
                            #if self._havePlayed:
                            if self._currentPlaylist and self._volume > 0:
                                try:
                                    self._app.stop()
                                    self._app.setSoundVolume_(self._origVolume)
                                except Exception,e: print 'Error stopping iTunes music:',e
                            self._currentPlaylist = None
                        else:
                            # if we've got something playing with positive volume, stop it
                            if self._currentPlaylist and self._volume > 0:
                                try:
                                    self._app.stop()
                                    self._app.setSoundVolume_(self._origVolume)
                                except Exception,e: print 'Error stopping iTunes music:',e
                            # set our playlist and play it if our volume is up
                            self._currentPlaylist = cmd[1]
                            if self._volume > 0:
                                self._origVolume = self._app.soundVolume()
                                self._updateITunesVolume()
                                self._playCurrentPlaylist()
                    elif cmd[0] == 'GET_PLAYLISTS':
                        try:
                            playlists = [p.name() for p in self._source.userPlaylists() if not p.name() in ['Music','Movies','TV Shows','Podcasts',u'iTunes\xa0U','Books','Genius','iTunes DJ', 'Music Videos']]
                            playlists.sort(key=lambda x: x.lower())
                        except Exception,e:
                            print 'Error getting iTunes playlists:',e
                            playlists = []
                        bs.runInGameThread(bs.Call(cmd[1],playlists))

                    del cmd # allows the command data/callback/etc to be freed

        def _playCurrentPlaylist(self):
            try:
                #print 'gonna do',self._currentPlaylist,type(self._currentPlaylist)
                playlistObj = self._source.userPlaylists().objectWithName_(self._currentPlaylist.decode('utf-8'))
                if playlistObj.exists():
                    # note: setting repeat and shuffle doesn't work in itunes 11 currently :-(
                    playlistObj.setSongRepeat_(1799449708) # repeat-all
                    playlistObj.setShuffle_(True)
                    playlistObj.playOnce_(False)
                else:
                    bs.runInGameThread(bs.WeakCall(bs.screenMessage,bs.getResource('playlistNotFoundText')+': \''+self._currentPlaylist+'\'',(1,0,0)))
            except Exception,e:
                try: print 'Exception playing playlist "'+self._currentPlaylist+'":',e
                except Exception: pass

        def _updateITunesVolume(self):
            self._app.setSoundVolume_(max(0,min(100,int(100.0*self._volume))))

        def setVolume(self,volume):
            oldVolume = self._volume
            self._volume = volume

            # if we've got nothing we're supposed to be playing, don't touch itunes..
            if self._currentPlaylist == None: return

            # if volume is going to zero, stop actually playing but dont clear playlist
            if oldVolume > 0 and volume == 0:
                try:
                    self._app.stop()
                    self._app.setSoundVolume_(self._origVolume)
                except Exception,e: print 'Error stopping iTunes music:',e
            elif self._volume > 0:
                # if volume was zero, store pre-playing volume and start playing
                if oldVolume == 0: self._origVolume = self._app.soundVolume()
                self._updateITunesVolume()
                if oldVolume == 0: self._playCurrentPlaylist()

        def playPlaylist(self,musicType):
            self._commands.append(['PLAY',musicType])
            self._commandsAvailable.set()

        def shutdown(self):
            self._commands.append(['DIE'])
            self._commandsAvailable.set()
            self.join()

        def getPlaylists(self,callback):
            self._commands.append(['GET_PLAYLISTS',callback])
            self._commandsAvailable.set()


    # define a music-player that talks to iTunes and lets us
    # select playlists as entries
    class MacITunesMusicPlayer(MusicPlayer):
        def __init__(self):
            MusicPlayer.__init__(self)
            self._thread = ITunesThread()
            self._thread.start()

        def onSelectEntry(self,callback,currentEntry,selectionTargetName):
            import bsUI
            return bsUI.GetSoundtrackEntryTypeWindow(callback,currentEntry,selectionTargetName)

        def onSetVolume(self,volume):
            self._thread.setVolume(volume)

        def onPlay(self,entry):
            entryType = _getSoundtrackEntryType(entry)
            if entryType == 'iTunesPlaylist':
                self._thread.playPlaylist(_getSoundtrackEntryName(entry))
            else:
                print 'MacITunesMusicPlayer passed unrecognized entry type:',entryType

        def onStop(self):
            self._thread.playPlaylist(None)

        def onShutdown(self):
            self._thread.shutdown()

    _gMusicPlayerType = MacITunesMusicPlayer


def haveMusicPlayer():
    return (_gMusicPlayerType is not None)

def getMusicPlayer():
    global _gMusicPlayer
    if _gMusicPlayer is None:
        if _gMusicPlayerType is None: raise Exception("no music player type set")
        _gMusicPlayer = _gMusicPlayerType()
    return _gMusicPlayer

# def getITunesThread():
#     global _gITunesThread
#     if _gITunesThread == None:
#         volume = _bs._getSetting('Music Volume')
#         print 'CREATING WITH VOLUME',volume
#         _gITunesThread = ITunesThread(volume=volume)
#         _gITunesThread.start()
#     return _gITunesThread
    
def _musicVolumeChanged(val):
    if _gMusicPlayer is not None:
        _gMusicPlayer.setVolume(val)
        
    # if _gITunesThread is not None:
    #     print 'SETTING VOLUME',val
    #     _gITunesThread.setVolume(val)

gMusicMode = 'regular'
gMusicTypes = {'regular':None,'test':None}

def setMusicPlayMode(mode,forceRestart=False):
    global gMusicMode
    oldMode = gMusicMode
    gMusicMode = mode
    if oldMode != gMusicMode or forceRestart:
        # if we're switching into test mode we don't
        # actually play anything until its requested..
        # if we're switching *out* of test mode though
        # we want to go back to whatever the normal song was
        if mode == 'regular': playMusic(gMusicTypes['regular'])


def _isSoundtrackEntryTypeSupported(entryType):
    ua = bs.getEnvironment()['userAgentString']
    if entryType == 'iTunesPlaylist':
        return True if 'Mac' in ua else False
    elif entryType in ('musicFile','musicFolder'):
        return True if ('android' in ua and _bs._androidGetExternalStoragePath() is not None) else False
    elif entryType == 'default':
        return True
        
def _getSoundtrackEntryType(entry):
    'Given a soundtrack entry, returns its type, taking into account what is supported locally.'
    try:
        if entry is None: entryType = 'default'
        # simple string denotes iTunesPlaylist (legacy format)
        elif type(entry) is str: entryType = 'iTunesPlaylist'
        # for other entries we expect type and name strings in a dict
        elif (type(entry) is dict
              and 'type' in entry and type(entry['type']) is str 
              and 'name' in entry and type(entry['name']) is str): entryType = entry['type']
        else: raise Exception("invalid soundtrack entry: "+str(entry))
        if _isSoundtrackEntryTypeSupported(entryType): return entryType
        else: raise Exception("invalid soundtrack entry:"+str(entry))
    except Exception,e:
        print 'EXC on _getSoundtrackEntryType',e
        return 'default'

def _getSoundtrackEntryName(entry):
    'Given a soundtrack entry, returns its name. Note that this does not ensure the entry is valid (use _getSoundtrackEntryType() for that).'
    try:
        if entry is None: return None
        # simple string denotes an iTunesPlaylist name (legacy format)
        elif type(entry) is str: return entry
        # for other entries we expect type and name strings in a dict
        elif (type(entry) is dict
              and 'type' in entry and type(entry['type']) is str 
              and 'name' in entry and type(entry['name']) is str): return entry['name']
        else: raise Exception("invalid soundtrack entry:"+str(entry))
    except Exception,e:
        print 'EXC on _getSoundtrackEntryName',e
        return 'default'

def playMusic(musicType,mode='regular',testSoundtrack=None):
    "Centralized func to change or stop the current music (pass None to stop)"

    gMusicTypes[mode] = musicType
    
    # if the OS tells us there's currently music playing, all our operations default to playing nothing
    if _bs._isOSPlayingMusic(): musicType = None

    # if we're not in the mode this music is being set for, don't actually change what's playing
    if mode != gMusicMode: return

    # some platforms have a special music-player for things like iTunes soundtracks, mp3s, etc.
    # if this is the case, attempt to grab an entry for this music-type, and
    # if we have one, have the music-player play it.  If not, we'll play game music ourself
    if musicType is not None and _gMusicPlayerType is not None:
        try:
            if testSoundtrack is not None:soundtrack = testSoundtrack
            else: soundtrack = bs.config['Soundtracks'][bs.config['Soundtrack']]
            entry = soundtrack[musicType]
        except Exception: entry = None
    else: entry = None

    # go through music-player
    if entry is not None:

        # stop any existing internal music..
        global gMusic
        if gMusic is not None:
            gMusic.delete()
            gMusic = None

        # play music-player music..
        getMusicPlayer().play(entry)
        
    # handle via internal music...
    else:
        if musicType is not None:
            loop = True
            if musicType == 'Menu':
                filename = 'menuMusic'
                volume = 5.0
            elif musicType == 'Victory':
                filename = 'victoryMusic'
                volume = 6.0
                loop = False
            elif musicType == 'CharSelect':
                filename = 'charSelectMusic'
                volume = 2.0
            elif musicType == 'RunAway':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Onslaught':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Keep Away':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Race':
                filename = 'runAwayMusic'
                volume = 6.0
            elif musicType == 'Epic Race':
                filename = 'slowEpicMusic'
                volume = 6.0
            elif musicType == 'Scores':
                filename = 'scoresEpicMusic'
                volume = 3.0
                loop = False
            elif musicType == 'GrandRomp':
                filename = 'grandRompMusic'
                volume = 6.0
            elif musicType == 'ToTheDeath':
                filename = 'toTheDeathMusic'
                volume = 6.0
            elif musicType == 'Chosen One':
                filename = 'survivalMusic'
                volume = 4.0
            elif musicType == 'ForwardMarch':
                filename = 'forwardMarchMusic'
                volume = 4.0
            elif musicType == 'FlagCatcher':
                filename = 'flagCatcherMusic'
                volume = 6.0
            elif musicType == 'Survival':
                filename = 'survivalMusic'
                volume = 4.0
            elif musicType == 'Epic':
                filename = 'slowEpicMusic'
                volume = 6.0
            elif musicType == 'Sports':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Hockey':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Football':
                filename = 'sportsMusic'
                volume = 4.0
            elif musicType == 'Flying':
                filename = 'flyingMusic'
                volume = 4.0
            elif musicType == 'Scary':
                filename = 'scaryMusic'
                volume = 4.0
            elif musicType == 'Marching':
                filename = 'whenJohnnyComesMarchingHomeMusic'
                volume = 4.0
            else:
                print "Unknown music: '"+musicType+"'"
                filename = 'flagCatcherMusic'
                volume = 6.0

        # stop any existing music-player playback
        if _gMusicPlayer is not None:
            _gMusicPlayer.stop()

        # stop any existing internal music..
        if gMusic is not None:
            gMusic.delete()
            gMusic = None

        # start up new internal music..
        if musicType is not None:
            # fixme - currently this won't start playing if we're paused
            # since attr values don't get updated until
            # node updates happen  :-(
            gMusic = bs.newNode(type='sound',
                                name='Global Music',
                                owner=bs,
                                attrs={'sound':bs.getSound(filename),
                                       'positional':False,
                                       'music':True,
                                       'volume':volume,
                                       'loop':loop})

class Spawner(object):
    """ given a delegate and some data, creates a light flash and a 'spawn' event.
    Note that the data is stored as-provided for the Spawner's duration,
    so pass it as a weak-ref is you need it to be able to die. """
    class SpawnMessage(object):
        def __init__(self,spawner,data,pt):
            self.spawner = spawner
            self.data = data
            self.pt = pt

    def __init__(self,activity,data,pt,spawnTime):

        self._spawnerSound = bs.getSound('swip2')

        self._activity = weakref.ref(activity)
        self._data = data
        self._pt = pt

        # create a light where the spawn will happen
        self._light = bs.newNode('light',owner=activity,attrs={'translate':tuple(pt),'radius':0.1,'color':(1.0,0.1,0.1),'lightVolumes':False})

        s = float(spawnTime)/3750
        minVal = 0.4
        maxVal = 0.7
        bs.playSound(self._spawnerSound,position=(self._light,'translate'))
        animate(self._light,'intensity',{0:0.0,
                                         int(250*s):maxVal,      int(500*s):minVal,  int(750*s):maxVal,      int(1000*s):minVal,
                                         int(1250*s):1.1*maxVal, int(1500*s):minVal, int(1750*s):1.2*maxVal, int(2000*s):minVal,
                                         int(2250*s):1.3*maxVal, int(2500*s):minVal, int(2750*s):1.4*maxVal, int(3000*s):minVal,
                                         int(3250*s):1.5*maxVal, int(3500*s):minVal, int(3750*s):2.0,        int(4000*s):0.0})
        bs.gameTimer(int(spawnTime),self._spawn)

    def _spawn(self):

        bs.gameTimer(1000,self._light.delete)

        # only run if our activity still exists
        activity = self._activity()
        if activity: activity.handleMessage(self.SpawnMessage(self,self._data,self._pt))

def isPointInBox(p,b):
    'for use with standard def boxes (translate|rotate|scale)'
    return ((abs(p[0]-b[0]) <= b[6]*0.5) and (abs(p[1]-b[1]) <= b[7]*0.5) and (abs(p[2]-b[2]) <= b[8]*0.5))

def getPlayerProfileColors(profileName):
    """ given a profile, returns colors for them """

    try: color = bs.config['Player Profiles'][profileName]['color']
    except Exception:
        # key off name if possible
        if profileName is None: 
            color = _gPlayerColors[random.randrange(6)] # first 6 are bright-ish
        else:
            color = _gPlayerColors[sum([ord(c) for c in profileName]) % 6] # first 6 are bright-ish

    try: highlight = bs.config['Player Profiles'][profileName]['highlight']
    except Exception:
        # key off name if possible
        if profileName is None: 
            highlight = _gPlayerColors[random.randrange(len(_gPlayerColors))]
        else:
            highlight = _gPlayerColors[sum([ord(c)+1 for c in profileName]) % len(_gPlayerColors)]

    return color,highlight

def getTimeString(t,centi=True):
    "Given a value in milliseconds, returns a string with minutes:seconds:centiseconds"
    m = str((t/1000)/60)
    s = str((t/1000)%60)
    cs = str(((t/10)%100))
    if len(s) == 1: s = '0'+s
    if len(cs) == 1: cs = '0'+cs
    if centi: return m+":"+s+"."+cs
    else: return m+":"+s

def showDamageCount(activity,damage,position,direction):
    lifespan = 1000
    t = bs.newNode('text',
                   owner=activity,
                   attrs={'text':damage,
                          'inWorld':True,
                          'hAlign':'center',
                          'color':(1,0.2,0.2,1),
                          'scale':[0.01]})
    # translate upward
    tcombine = bs.newNode("combine",owner=t)
    tcombine.connectAttr('output',t,'translate')
    vVals = []
    p = 0; v = 0.07
    count = 6
    for i in range(count):
        vVals.append((float(i)/count,p))
        p += v
        v *= 0.5
    pStart = position[0]; pDir = direction[0]
    animate(tcombine,"input0",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))

    pStart = position[1]; pDir = direction[1]
    animate(tcombine,"input1",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))

    pStart = position[2]; pDir = direction[2]
    animate(tcombine,"input2",dict([[i[0]*lifespan,pStart+pDir*i[1]] for i in vVals]))
    ccombine = bs.newNode("combine",owner=t,attrs={'input0':1.0,'input1':0.4,'input2':0.4})

    ccombine.connectAttr('output',t,'color')
    animate(ccombine,"input3",{0.7*lifespan:1.0,lifespan:0.0})
    bs.gameTimer(lifespan,t.delete)

def getLastPlayerNameFromInputDevice(device):
    'Returns a reasonable player name associated with a device (generally the last one used there)'
    # look for a default player profile name for them.. otherwise default to their current random name
    profileName = '_random'
    keyName = device.getName()+' '+device.getUniqueIdentifier()
    if 'Default Player Profiles' in bs.config and keyName in bs.config['Default Player Profiles']:
        profileName = bs.config['Default Player Profiles'][keyName]
    if profileName == '_random': profileName = device._getDefaultPlayerName()
    return profileName

def getNormalizedColor(color):
    """ scale a color so its largest value is 1; useful for coloring lights"""
    colorBiased = tuple(max(c,0.01) for c in color) # account for black and whatnot
    mult = 1.0/max(colorBiased)
    return tuple(c*mult for c in colorBiased)


def getSafeDisplayColor(color):
    """ given a color, returns one that is safe to display over black """

    c = bs.getSafeColor(color=(color[0],color[1],color[2]));
    if len(color) == 4: return c+(float(color[3]),)
    else: return c


def garbageCollect():
    gc.collect()
    # can be handy to print this to check for leaks between games
    if 0: print 'PY OBJ COUNT',len(gc.get_objects())
    if len(gc.garbage) > 0:
        print 'PTHON GC FOUND',len(gc.garbage),'UNCOLLECTABLE OBJECTS:'
        for i,obj in enumerate(gc.garbage):
            print str(i)+':',obj

    printLiveObjectWarnings('after session shutdown')

def printLiveObjectWarnings(when,ignoreSession=None,ignoreActivity=None):
    sessions = []
    activities = []
    actors = []
    materials = []
    global _gPrintedLiveObjectWarning
    
    if _gPrintedLiveObjectWarning:
        # print 'skipping live object check due to previous found live object(s)'
        return

    for obj in gc.get_objects():
        if isinstance(obj,bs.Actor): actors.append(obj)
        elif isinstance(obj,bs.Session): sessions.append(obj)
        elif isinstance(obj,bs.Activity): activities.append(obj)
        elif isinstance(obj,bs.Material): materials.append(obj)

    # complain about any remaining sessions
    for session in sessions:
        if session is ignoreSession: continue
        _gPrintedLiveObjectWarning = True
        print 'Session found',when,':',type(session)
        refs = list(gc.get_referrers(session))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining activities
    for activity in activities:
        if activity is ignoreActivity: continue
        _gPrintedLiveObjectWarning = True
        print 'Activity found',when,':',type(activity)
        refs = list(gc.get_referrers(activity))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining actors
    for actor in actors:
        _gPrintedLiveObjectWarning = True
        print 'Actor found',when,':',type(actor)
        if type(actor) is bs.Actor:
            try:
                if actor.node.exists(): print '   - contains node:',actor.node.getType(),';',actor.node.getName()
            except Exception,e:
                print '   - exception checking actor node:',e
        refs = list(gc.get_referrers(actor))
        i = 1
        for ref in refs:
            if type(ref) is types.FrameType: continue
            print '     ref',i,':',ref
            i += 1
            if type(ref) is list or type(ref) is tuple or type(ref) is dict:
                refs2 = list(gc.get_referrers(ref))
                j = 1
                for ref2 in refs2:
                    if type(ref2) is types.FrameType: continue
                    print '        ref\'s ref',j,':',ref2
                    j += 1

    # complain about any remaining materials
    for material in materials:
        _gPrintedLiveObjectWarning = True
        print 'Material found',when,':',material.getName()
                    
    # complain about any remaining nodes
    j = 1
    for node in bs.getNodes():
        if _isValidRemainingNode(node): continue
        _gPrintedLiveObjectWarning = True
        print 'Live node found',when,':',j,':',node.getType(),';',node.getName()
        j += 1

def _isValidRemainingNode(node):
    return (node.getName() in ['Globals','Global Music','Global Time'])

def _weakDeleteNode(node,weakRef):
    node.delete()

def _setNodeOwner(node,owner,isCrossFade):

    # ..owner can be an activity..
    if isinstance(owner,bs.Activity):
        owner._addOwnedNode(node,isCrossFade)

    # a few key nodes need to never get auto-deleted; they pass bs in this case
    # (hmm i suppose the weak-ref case would cover this too..)
    elif owner is bs:
        pass

    # for all other types, create a weak-ref callback to kill the node when the target object dies
    # hmm should we explicitly limit this to actors or something?
    else:
        global gNodeOwnerWeakRefs
        global gNodeOwnerWeakRefsCleanCounter
        gNodeOwnerWeakRefs.append(weakref.ref(owner,bs.Call(_weakDeleteNode,node)))
        gNodeOwnerWeakRefsCleanCounter += 1
        # pull dead refs off our list periodically (I supposed the callback could do it, but this works)
        if (gNodeOwnerWeakRefsCleanCounter > 10):
            gNodeOwnerWeakRefs = [r for r in gNodeOwnerWeakRefs if r() is not None]
            gNodeOwnerWeakRefsCleanCounter = 0
    
class ServerCallThread(threading.Thread):
    def __init__(self,request,requestType,data,callback):
        threading.Thread.__init__(self)
        self._request = request
        self._requestType = requestType
        self._data = {} if data == None else data
        self._callback = callback

    def run(self):
        try:
            if self._requestType == 'get':
                response = urllib2.urlopen(urllib2.Request(_bs._getServerAddress()+'/'+self._request+'?'+urllib.urlencode(self._data),
                                                           None,
                                                           { 'User-Agent' : bs.getEnvironment()['userAgentString'] }))
            elif self._requestType == 'post':
                response = urllib2.urlopen(urllib2.Request(_bs._getServerAddress()+'/'+self._request,
                                                           urllib.urlencode(self._data),
                                                           { 'User-Agent' : bs.getEnvironment()['userAgentString'] }))
            else: raise Exception("Invalid requestType: "+self._requestType)
            responseData = ast.literal_eval(response.read())
            if self._callback is not None: bs.runInGameThread(bs.Call(self._callback,responseData))
        except Exception,e:
            if self._callback is not None: bs.runInGameThread(bs.Call(self._callback,None))


def serverGet(request,data,callback=None):
    ServerCallThread(request,'get',data,callback).start()

def serverPut(request,data,callback=None):
    ServerCallThread(request,'post',data,callback).start()

def runStressTest(playlistType='Random',playlistName='__default__',playerCount=8,roundDuration=120):
    bs.screenMessage('Beginning stress test.. use \'End Game\' to stop testing.',color=(1,1,0))
    bs.realTimer(1000,bs.WeakCall(_startStressTest,{'playlistType':playlistType,
                                                    'playlistName':playlistName,
                                                    'playerCount':playerCount,
                                                    'roundDuration':roundDuration}))
    bs.realTimer(8000,bs.WeakCall(bs.screenMessage,'stress-test stats will be written to: '+bs.getEnvironment()['userScriptsDirectory']+'/stressTestStats.csv'))

def stopStressTest():
    _bs._setStressTesting(False,0)
    global _gStressTestResetTimer
    try:
        if _gStressTestResetTimer is not None:
            bs.screenMessage("Ending stress test...",color=(1,1,0))
    except Exception: pass
    _gStressTestResetTimer = None

def _startStressTest(args):
    playlistType = args['playlistType']
    if playlistType == 'Random':
        if (random.random() < 0.5): playlistType = 'Teams'
        else: playlistType = 'Free-For-All'
    bs.screenMessage('Running Stress Test (listType="'+playlistType+'", listName="'+args['playlistName']+'")...')
    if playlistType == 'Teams':
        bs.config['Team Tournament Playlist Selection'] = args['playlistName']
        bs.config['Team Tournament Playlist Randomize'] = 1
        _bs._hostSession('bsTeamGame.TeamsSession')
    else:
        bs.config['Free-for-All Playlist Selection'] = args['playlistName']
        bs.config['Free-for-All Playlist Randomize'] = 1
        _bs._hostSession('bsTeamGame.FreeForAllSession')
    _bs._setStressTesting(True,args['playerCount'])
    global _gStressTestResetTimer
    _gStressTestResetTimer = bs.Timer(args['roundDuration']*1000,bs.WeakCall(_resetStressTest,args),timeType='real')


def _resetStressTest(args):
    _bs._setStressTesting(False,args['playerCount'])
    bs.screenMessage('Resetting stress test...')
    bs.getSession().end()
    bs.realTimer(2000,bs.WeakCall(_startStressTest,args))
